/*
 * client.c
 * client side of superserver
 * 
 * Usage: client [-p port] <hostname> <service> [args]
 *
 * This is the "client side" of the superserver utility suite.  See README
 * for details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/errno.h>
#include <sys/signal.h>

#include "socketlib.h"
#include "config.h"

extern int errno;

static void usage(), reap(), inout(int);

main(int argc, char **argv)
{
    extern char *optarg;
    extern int optind;
    int port = DEFAULT_PORT, sock, flags;
    char *hostname, *service;
    char command[BUFSIZ], buffer[BUFSIZ];
    char c, *end;

    /* parse command line options */
    while((c = getopt(argc, argv, "p:")) != -1) {
	if(c == 'p') {
	    port = strtol(optarg, &end, 10);
	    if(optarg == end) {
		fprintf(stderr,
		        "FATAL ERROR: invalid port number %s\n", optarg);
		exit(1);
	    }
	    break;
        } else {
	    usage();
	}
    }

    if(argc - optind < 2) {
	usage();
    }

    hostname = argv[optind++];
    service = argv[optind++];

    strncpy(command, service, BUFSIZ - 1);
    for(; optind < argc; optind++) {
	strncat(command, " ", BUFSIZ - 1);
	strncat(command, argv[optind], BUFSIZ - 1);
    }
    strncat(command, "\n", BUFSIZ - 1);

    if((sock = clientsock(hostname, port)) < 0) {
	fprintf(stderr, "FATAL ERROR: Could not connect to server.\n");
	exit(1);
    }

    if(write(sock, command, strlen(command)) < 0) {
	perror("write");
	exit(1);
    }

    inout(sock);
    exit(0);
}

/*
 * inout:
 * process input and output on socket s.
 *
 * inout() calls fork(), so that one process copies stdin to the socket,
 * and the other copies the socket to stdout; this allows both input and
 * output to run simultaneously.  when eof is reached on either side,
 * both processes terminate.
 */ 
static void
inout(int s)
{
    char buffer[BUFSIZ];
    int pid, n;

    signal(SIGCLD, reap);

    if((pid = fork()) != 0) {
	close(0);
	while((n = read(s, buffer, BUFSIZ)) > 0) {
	    write(1, buffer, n);
	}
	close(s);
	kill(pid, SIGKILL);
    } else {
	close(1);
        while((n = read(0, buffer, BUFSIZ)) > 0) {
	    write(s, buffer, n);
        }
	close(s);
    }
}

/*
 * reap:
 * pick up child processes that have died
 */
static void
reap()
{
    wait(NULL);
    exit(0);
}

/*
 * usage:
 * print usage message and exit
 */
static void
usage()
{
    fprintf(stderr, "Usage: client [-p port] <hostname> <service> [args]\n");
    exit(1);
}

