/*
 * rcfile.c
 * routines for reading and writing the service DB file for the
 * superserver utility suite
 */

#include <stdio.h>
#include <stdlib.h>
#include "sdb.h"
#include "rcfile.h"
#include "config.h"

static const char *rcfile = RCFILE;

/*
 * nextf:
 * skip to the next field in a colon-deliminated line of text
 */
static char *
nextf(register char *p)
{
    while(*p && *p != '\n' && *p != ':') {
	p++;
    }

    if(*p == '\n') {
	*p = '\0';
    } else {
	*p++ = '\0';
    }

    return(p);
}
    
/*
 * rc_loadall:
 * read in the service DB file (usually /usr/local/bin/svcrc)
 *
 * TODO: need file locking, since sregister can modify the file while
 * supersrv is reading it
 */
int
rc_loadall()
{
    svcid_t svc;
    char *line, *p, *end;
    static int size = 256;
    int len;
    FILE *sfp;
    
    if((sfp = fopen(rcfile, "r")) == NULL) {
	return(0);
    }

    if((line = malloc(size)) == NULL) {
	perror("malloc");
	return(0);
    }

    /* read line into buffer */
    while(fgets(line, size, sfp) != NULL) {
        len = strlen(line);
        while(line[len-1] != '\n' && !feof(sfp)) {
            /* line is too long to fit; double buffer size and try again */
            size *= 2;
            line = realloc(line, size);
            if(line == NULL) {
                perror("realloc");
	        return(0);
            }

            if(fgets(&line[len], size/2, sfp) == NULL) {
                break;
            }

            len = strlen(line);
        }

        /* now split the line into fields: first, svc_name */
        p = line;
        svc.svc_name = p;

        /* svc_cmd */
        p = nextf(p);
	if(*p) {
	    svc.svc_cmd = p;
	} else {
	    fprintf(stderr, "%s: badly formatted line.\n", rcfile);
	}

        /* svc_owner */
	p = nextf(p);
	svc.svc_owner = strtol(p, &end, 10);
	if(p == end) {
	    fprintf(stderr, "%s: badly formatted line.\n", rcfile);
	    break;
	}

        /* svc_uid */
	p = nextf(p);
	svc.svc_uid = strtol(p, &end, 10);
	if(p == end) {
	    fprintf(stderr, "%s: badly formatted line.\n", rcfile);
	    break;
	}

        /* svc_gid */
	p = nextf(p);
	svc.svc_gid = strtol(p, &end, 10);
	if(p == end) {
	    fprintf(stderr, "%s: badly formatted line.\n", rcfile);
	    break;
	}

        /* load the service into the hash table */
	svc_register(&svc);
    }

    fclose(sfp);
    return(1);
}

/*
 * rc_saveall:
 * save the contents of the service DB to disk (normally /usr/local/lib/svcrc)
 *
 * TODO: need file locking, since sregister can modify the file while
 * supersrv is reading it
 */
int
rc_saveall()
{
    svcid_t *svcs, *s;
    int i, num_svcs;
    FILE *sfp;

    if((sfp = fopen(rcfile, "w")) == NULL) {
	perror("fopen");
	return(0);
    }

    svcs = svc_list(&num_svcs);
    for(i = 0; i < num_svcs; i++) {
	fprintf(sfp, "%s:%s:%d:%d:%d\n", svcs[i].svc_name,
					 svcs[i].svc_cmd,
					 svcs[i].svc_owner,
					 svcs[i].svc_uid,
					 svcs[i].svc_gid);
    }
    fclose(sfp);
    free(svcs);
}

