/*
 * sdb.c
 * routines to store and retrieve service information.
 *
 * implemented as a hash table of linked lists.
 */

#include <stdio.h>
#include <string.h>
#include "sdb.h"

#define HASH_SLOTS 256

svcid_t *SVCS[HASH_SLOTS];
int num_svcs;

/*
 * svc_init:
 * if SVCS table hasn't been initialized yet, set all the buckets to NULL.
 * this should be run at the beginning of every routine that accesses SVCS.
 */
static void
svc_init()
{
    static initialized = 0;
    int i;

    if(initialized) {
	return;
    }

    for(i = 0; i < HASH_SLOTS; i++) {
	SVCS[i] = NULL;
    }
    num_svcs = 0;
    initialized = 1;
}

/*
 * svc_free:
 * free the memory associated with a a service, and decrement num_svcs.
 */
static void
svc_free(svcid_t *svc)
{
    num_svcs--;

    if(svc != NULL) {
        if(svc->svc_cmd != NULL)
    	    free(svc->svc_cmd);
        if(svc->svc_name != NULL)
            free(svc->svc_name);
        free(svc);
    }
}

/*
 * svc_dup:
 * allocate memory and copy a server entry into it; increment num_svcs.
 * returns newly allocated svcid_t on success, or NULL on failure.
 */
static svcid_t *
svc_dup(svcid_t *svc)
{
    svcid_t *newsvc;

    num_svcs++;

    newsvc = (svcid_t *) malloc(sizeof(svcid_t));
    if(newsvc == NULL) {
	fprintf(stderr, "ERROR: Unable to register service %s\n",
	        svc->svc_name);
	perror("malloc");
	return(NULL);
    }

    newsvc->svc_name = strdup(svc->svc_name);
    if(newsvc->svc_name == NULL) {
	fprintf(stderr, "ERROR: Unable to register service %s\n",
	        svc->svc_name);
	perror("malloc");
	svc_free(newsvc);
	return(NULL);
    }

    newsvc->svc_cmd = strdup(svc->svc_cmd);
    if(newsvc->svc_cmd == NULL) {
	fprintf(stderr, "ERROR: Unable to register service %s\n",
	        svc->svc_name);
	perror("malloc");
	svc_free(newsvc);
	return(NULL);
    }

    newsvc->svc_owner = svc->svc_owner;
    newsvc->svc_uid = svc->svc_uid;
    newsvc->svc_gid = svc->svc_gid;
    newsvc->svc_next = NULL;

    return(newsvc);
}

/*
 * hash:
 * simple hash function for service names
 */
static int
hash(char *input)
{
    int sum = 0;

    while(*input) {
	sum += (int) *input++;
	sum %= HASH_SLOTS;
    }
    return(sum);
}

/*
 * svc_register:
 * register a new service.
 * input: a pointer to a service ID structure
 * returns: 1 on success, 0 on failure
 *
 * attempting to register a service already belonging to another user
 * is a fatal error and terminates the program (unless you're root).
 */
int
svc_register(svcid_t *svc)
{
    int index;
    svcid_t *s;

    svc_init();

    index = hash(svc->svc_name);

    if(SVCS[index] == NULL) {
	SVCS[index] = svc_dup(svc);
	return(SVCS[index] != NULL);
    }

    s = SVCS[index];
    for(;;) {
        if(strcmp(s->svc_name, svc->svc_name) == 0) {
	    /* found service with the same name; are we allowed to ovewrite? */
	    if(s->svc_owner != svc->svc_owner && getuid() != 0) {
                fprintf(stderr, "FATAL ERROR: Service '%s' already registered "
	    		        "by another user.\n", s->svc_name);
                exit(1);
            }

            /* yes we are */
	    free(s->svc_cmd);
	    s->svc_cmd = strdup(svc->svc_cmd);
	    s->svc_owner = svc->svc_owner;
	    s->svc_uid = svc->svc_uid;
	    s->svc_gid = svc->svc_gid;
	    break;
        }

        if(s->svc_next) {
    	    s = s->svc_next;
        } else {
            s->svc_next = svc_dup(svc);
	    return(s->svc_next != NULL);
        }
    }
}

/*
 * svc_unregister:
 * remove an existing service from the database
 *
 * input: name of the service
 * returns: 1 on success, 0 on failure
 *
 * attempting to delete a nonexistent service generates a warning message
 *
 * attempting to delete a service belonging to another user is a fatal
 * error and terminates the program (unless you're root).
 *
 * TODO: this is klugy; refactor to eliminate special-case code
 */
int
svc_unregister(char *svc_name)
{
    int index, uid;
    svcid_t *s, *t;

    svc_init();
    
    uid = getuid();

    index = hash(svc_name);
    s = SVCS[index];

    if(s != NULL) {
        /* special case: we found a hit in the list header */
        if(strcmp(s->svc_name, svc_name) == 0) {
	    if(uid && (s->svc_owner != uid)) {
                fprintf(stderr, "FATAL ERROR: Service '%s' is registered "
	    		        "by another user.\n", s->svc_name);
                exit(1);
            }

	    SVCS[index] = s->svc_next;
	    svc_free(s);
	    return(1);
        }

        while(s->svc_next != NULL) {
	    if(strcmp(s->svc_next->svc_name, svc_name) != 0) {
	        s = s->svc_next;
	        continue;
            }

	    if(uid && (s->svc_owner != uid)) {
                fprintf(stderr, "FATAL ERROR: Service '%s' is registered "
	    		        "by another user.\n", s->svc_name);
                exit(1);
            }

	    t = s->svc_next;
            s->svc_next = s->svc_next->svc_next;
	    svc_free(t);
	    return(1);
        }
    }

    /* if we're here, the service name was not found */
    fprintf(stderr, "ERROR: Service %s is not registered.\n", svc_name);
    return(0);
}

/*
 * svc_get:
 * return a pointer to a service ID structure matching a given service name
 *
 * input: service name
 * output: svcid_t pointer, or NULL on failure
 */
svcid_t *
svc_get(char *svc_name)
{
    svcid_t *s;

    svc_init();

    s = SVCS[hash(svc_name)];
    while(s != NULL) {
	if(strcmp(s->svc_name, svc_name) == 0) {
	    return(s);
	}
	s = s->svc_next;
    }

    /* if we're here, the service name was not found */
    return(NULL);
}

/*
 * svc_list:
 * provide a list of all services stored in the hash table, and the number
 *
 * input: pointer to an int in which to store the number of services
 * returns: array of svcid_t
 */
svcid_t *
svc_list(int *ns)
{
    svcid_t *svcs, *s, *t;
    int i;

    svc_init();

    svcs = (svcid_t *) malloc(sizeof(svcid_t) * num_svcs);
    s = svcs;

    for(i = 0; i < HASH_SLOTS; i++) {
	if((t = SVCS[i]) != NULL) {
	    bcopy(t, s, sizeof(svcid_t));
	    s++;
	    while(t->svc_next != NULL) {
		t = t->svc_next;
	        bcopy(t, s, sizeof(svcid_t));
	        s++;
            }
        }
    }

    *ns = num_svcs;
    return svcs;
}

/*
 * svc_clear:
 * empty the hash table, free all memory associated with existing services,
 * set all buckets to NULL
 */
void
svc_clear()
{
    svcid_t *s, *t;
    int i;

    svc_init();

    for(i = 0; i < HASH_SLOTS; i++) {
	s = SVCS[i];
        while(s != NULL) {
            t = s->svc_next;
            svc_free(s);
	    s = t;
        }
	SVCS[i] = NULL;
    }
}

#ifdef DEBUG
/* unit tests */
main()
{
    svcid_t svc, *s, *svcs;
    int i, ns;

    svc.svc_name = "who";
    svc.svc_cmd = "/bin/who";
    svc.svc_owner = svc.svc_uid = getuid();
    svc.svc_gid = getgid();
    svc_register(&svc);

    svc.svc_name = "ls";
    svc.svc_cmd = "/usr/bin/ls";
    svc_register(&svc);

    svc_clear();

    svc.svc_name = "how";
    svc.svc_cmd = "/usr/local/bin/how";
    svc_register(&svc);

    svcs = svc_list(&ns);
    for(i = 0; i < ns; i++) {
	s = svcs + i;
	printf("%s\t%s\n", s->svc_name, s->svc_cmd);
    }
    free(svcs);

    svc_unregister("how");
    svcs = svc_list(&ns);
    for(i = 0; i < ns; i++) {
	printf("%s\t%s\n", svcs[i].svc_name, svcs[i].svc_cmd);
    }
    free(svcs);
}
#endif
