/*
** Some blackbox socket routines
** (evanh@sco.com)
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>

static void pickup_child();

static int remote_port;
static long remote_addr;

extern int errno;

int debugmode = 0;

/*
** serversock -- open a socket and bind to a port address (leaving
**               listen/accept to calling function)
**
** Input: port number of socket, or 0 for a random port.
** Output: file descriptor of connected socket, or -errno.
*/

serversock(port)
int port;
{
    struct sockaddr_in sin;
    long rhost();
    int sock;

    sin.sin_family = AF_INET;
    sin.sin_port = htons(port);
    sin.sin_addr.s_addr = INADDR_ANY;

    if((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
        perror("socket");    
        return(-errno);
    }

    if(bind(sock, (struct sockaddr *) &sin, sizeof(sin)) == -1) {
        perror("accept");
        return(-errno);
    }

    return(sock);
}

/*
** servrun -- listen for and accept connections; fork child processes to
**            wait on htem
**
** Input: pointer to function which will be called when a connection
**        arrives, and a port number to listen on.  The function should
**        take as input the file descriptor.
** Output: none
*/

servrun(callfunc, port)
int (*callfunc)(), port;
{
    struct sockaddr_in rsin;
    int s, ns, size = sizeof(rsin);

    s = serversock(port);

    if(listen(s, 5) < 0) {
        perror("listen");
        return(-errno);
    }

    for(;;) {
        if((ns = accept(s, (struct sockaddr *) &rsin, &size)) < 0) { 
            if(errno == EINTR) {
                continue;
            } else {
                perror("accept");
                return(-errno);
            }
        }

        /* save information about client for possible future reference */
        remote_port = ntohs(rsin.sin_port);
        remote_addr = ntohs(rsin.sin_addr.s_addr);

        if(!fork()) {
	    if(debugmode) {
		fprintf(stderr, "Connection received\n");
            }
            (*callfunc)(ns);
            exit(0);
        } else {
            signal(SIGCLD, pickup_child);
            close(ns);
        }
    }
}

/*
** clientsock -- connect to a waiting socket and return a file descriptor.
**
** Input: host name and port number to connect to.
** Output: file descriptor of connected socket or a negative error.
*/

clientsock(host, port)
char *host;
int port;
{
    int s;
    struct sockaddr_in sin;
    struct hostent *hp;

    if ((hp = gethostbyname(host)) == 0) {
	perror("gethostbyname");
        return(-errno);
    }

    bzero(&sin, sizeof(sin));
    bcopy((char *) hp->h_addr, (char *) &sin.sin_addr.s_addr, hp->h_length);
    sin.sin_family = AF_INET;
    sin.sin_port = htons(port);

    if ((s = socket(sin.sin_family, SOCK_STREAM, 0)) == -1) {
        perror("socket");
        return(-errno);
    }

    if (connect(s, (struct sockaddr *) &sin, sizeof(sin)) == -1) {
        perror("connect");
        return(-errno);
    }

    return(s);
}

/*
** backgrd -- prepare calling process for functioning as a daemon.
**
** Drop into background, disassociate from tty, ignore 
** termination signals, and point standard files at /dev/null.
** NOTE:  If application uses signal traps, they should be
** set AFTER this function is called, because it will
** nullify signals 1, 2, 3 and 15.
**
** No input or output.
*/

backgrd()
{

    signal(SIGHUP, SIG_IGN);
    signal(SIGINT, SIG_IGN);
    signal(SIGQUIT, SIG_IGN);

    close(0), close(1), close(2);
    open("/dev/null", O_RDONLY);
    open("/dev/null", O_WRONLY);
    dup(1);

    setpgrp();

    if(fork())
        exit(0);
}

/*
** Pick up a dead child process.
*/

static void reset_pickup()
{
    signal(SIGCLD, pickup_child);
}

void pickup_child() 
{
    wait(NULL);
    reset_pickup();
}
