/*
 * sregister
 * configuration tool for superserver utility suite.
 *
 * Registers or removes services from service DB, then signals the server
 * that the services have been updated
 *
 * Usage:
 *   sregister [-o user|uid] [-u user|uid] [-g group|gid] service command
 *   sregister -d service
 *   sregister -l
 *
 * -o option is only available for root; it sets the service owner to a
 *    particular user, who can then change or delete it
 *
 * -u option is only available for root; by default, always "nouser".  sets
 *    the UID to use for running the service
 *
 * -g option is only available for root; by default, always "nogroup".  sets
 *    the GID to use for running the service
 *
 * -d option is unlimited for root; for other users, limited to services
 *    owned by the user
 *
 * -l lists all currently registered services
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/signal.h>
#include <pwd.h>
#include <grp.h>
#include "sdb.h"
#include "rcfile.h"
#include "config.h"

static int list_services();
static int usage();

main(int argc, char **argv)
{
    svcid_t svc;
    struct passwd *pw;
    struct group *gr;
    int list = 0, delete = 0;
    extern char *optarg;
    extern int optind;
    char c, *end, line[BUFSIZ];
    uid_t uid;
    int pid;
    FILE *fp;

    /* set default values for owner, uid, gid */
    if(pw = getpwnam("nouser")) {
       svc.svc_uid = pw->pw_uid;
       svc.svc_gid = pw->pw_gid;
    } else {
        fprintf(stderr, "FATAL ERROR: Couldn't find user 'nouser'.\n");
	exit(1);
    }

    uid = getuid();
    svc.svc_owner = uid;

    /* parse command line args */
    while((c = getopt(argc, argv, "ldo:u:g:")) != -1) {
	switch(c) {
	    case 'd':
		delete = 1;
		break;
	    case 'l':
		list = 1;
		break;
	    case 'o':
		if(uid) {
		    usage();
		}

		if(pw = getpwnam(optarg)) {
		    svc.svc_owner = pw->pw_uid;
                } else {
		    svc.svc_owner = strtol(optarg, &end, 10);
		    if(optarg == end) {
			usage();
		    }
		}

		break;
	    case 'u':
		if(uid) {
		    usage();
		}

		if(pw = getpwnam(optarg)) {
		    svc.svc_uid = pw->pw_uid;
                } else {
		    svc.svc_uid = strtol(optarg, &end, 10);
		    if(optarg == end) {
			usage();
		    }
		}

		break;
	    case 'g':
		if(uid) {
		    usage();
		}

		if(gr = getgrnam(optarg)) {
		    svc.svc_gid = gr->gr_gid;
                } else {
		    svc.svc_gid = strtol(optarg, &end, 10);
		    if(optarg == end) {
			usage();
		    }
		}

		break;
            case '?':
		usage();
		break;
	}
    }

    /* load service DB */
    rc_loadall();

    if(list) {
        /* list all services */
	if(argc != optind) {
	    usage();
        }

	list_services();
	exit(0);
    }

    if(delete) {
   	 /* delete a service */
	if((argc - optind) != 1) {
	    usage();
        }
        
	svc_unregister(argv[optind]);
    } else {
        /* add or modify a service */
        if(argc - optind != 2) {
            usage();
        }

        svc.svc_name = argv[optind];
        svc.svc_cmd = argv[optind + 1];

        svc_register(&svc);
    }

    /* we're done making changes; write the service DB back out to disk */
    rc_saveall();

    /* signal the server that it needs to reload. */
    if((fp = fopen(PIDFILE, "r")) == NULL || fgets(line, BUFSIZ, fp) == NULL) {
	/* can't read the pid?  okay. */
	exit(0);
    }

    fclose(fp);

    /* if the first line of PIDFILE is a pid, send it a SIGHUP */
    pid = strtol(line, &end, 10);
    if(line != end) {
	kill(pid, SIGHUP);
    }

    exit(0);
}

/*
 * list_services:
 * print a list of all services currently registered
 */
static int
list_services()
{
    svcid_t *svcs;
    struct passwd *pw;
    int nsvcs, i;
    char *owner;

    printf("service\tcommand\towner\tuid\tgid\n");
    printf("-------\t-------\t-----\t---\t---\n");

    svcs = svc_list(&nsvcs);
    for(i = 0; i < nsvcs; i++) {
	if(pw = getpwuid(svcs[i].svc_owner)) {
	    owner = pw->pw_name;
	    printf("%s\t%s\t%s\t%d\t%d\n", svcs[i].svc_name,
					   svcs[i].svc_cmd,
					   owner,
					   svcs[i].svc_uid,
					   svcs[i].svc_gid);
	} else {
	    printf("%s\t%s\t%d\t%d\t%d\n", svcs[i].svc_name,
					   svcs[i].svc_cmd,
					   svcs[i].svc_owner,
					   svcs[i].svc_uid,
					   svcs[i].svc_gid);
	}
    }

}

/*
 * usage:
 * print usage message and exit
 */
static int
usage()
{
    if(getuid() == 0) {
        fprintf(stderr, "Usage:\tsregister [-u <user|uid>] [-g <group|gid>] service command\n");
    } else {
        fprintf(stderr, "Usage:\tsregister service command\n");
    }

    fprintf(stderr, "\tsregister -d service\n");
    fprintf(stderr, "\tsregister -l\n");

    exit(1);
}
