/*
 * supersrv.c
 *
 * Usage: supersrv [-D] [-p port]
 *
 * The server side of the superserver utility suite; a generic TCP/IP
 * service provider.  See README for details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/signal.h>
#include "sdb.h"
#include "rcfile.h"
#include "socketlib.h"
#include "config.h"

extern int debugmode;

static int superserver(), reload();
static void list_services();
static char **getargs(char *);

main(int argc, char **argv)
{
    long port = DEFAULT_PORT;
    extern char *optarg;
    char c, *end;
    FILE *fp;

    /* parse command line args */
    while((c = getopt(argc, argv, "Dp:")) != -1) {
	switch(c) {
	    case 'D':
	        debugmode = 1;
	        break;
            case 'p':
		port = strtol(optarg, &end, 10);
		if(optarg == end) {
		    fprintf(stderr, "FATAL ERROR: invalid port number %s\n", optarg);
		    exit(1);
		}
		break;
	    case '?':
	        fprintf(stderr, "Usage: %s [-D]\n", argv[0]);
		exit(1);
		break;
	}
    }

    /* run in the background */
    if(!debugmode) {
        backgrd();
    }

    /* save PID in PIDFILE */
    if((fp = fopen(PIDFILE, "w")) != NULL) {
	fprintf(fp, "%d\n", getpid());
	fclose(fp);
    }

    /* load service DB */
    rc_loadall();
    signal(SIGHUP, reload);

    /* begin offering services now */
    if(debugmode) {
	fprintf(stderr, "Listening for connections on port %d\n", port);
    }
    servrun(superserver, port);
}

/*
 * superserver:
 * called whenever a new connection arrives, this routine delivers services
 * to clients.  
 */
static int
superserver(int s)
{
    char cmd[BUFSIZ], **argv;
    svcid_t *svc;
    int n;

    signal(SIGHUP, SIG_DFL);

    /*
     * if we're in debug mode, then stdio hasn't been closed yet; 
     * we need to do that now.
     */
    if(debugmode) {
        close(0);
        close(1);
        close(2);
    }

    if(dup2(s, 0) != 0 || dup2(s, 1) != 1 || dup2(s, 2) != 2) {
	/* unable to tie socket to stdio */
	exit(1);
    }

    if((n = read(0, cmd, BUFSIZ)) <= 0) {
	/* read failed, eof or error: either way, we're done */
	exit(1);
    }

    if(n == BUFSIZ) {
        /* buffer overrun: just close the connection */
        exit(0);
    }

    /* build argument vector from input */
    argv = getargs(cmd);

    if(strcmp(argv[0], "list") == 0) {
	list_services();
    }

    /* does the requested service exist? */
    if((svc = svc_get(argv[0])) == NULL) {
        printf("FATAL ERROR: Service '%s' is not registered.\n", argv[0]);
	exit(1);
    }

    /* set user and group */
    setuid(svc->svc_uid);
    setgid(svc->svc_gid);

    /* run the server command */
    argv[0] = svc->svc_cmd;
    execv(argv[0], argv + 1);

    /* this should never be reached */
    exit(1);
}

/*
 * list_services:
 * called when the client requests "list"; prints a simple list of
 * all services registered.
 *
 * TODO: would be nice to supply a short description as well
 */
static void
list_services()
{
    svcid_t *svcs;
    int i, nsvcs;

    svcs = svc_list(&nsvcs);
    if(!nsvcs) {
        printf("No services registered on this host.\n");
	exit(0);
    }

    printf("Services registered on this host:\n");
    for(i = 0; i < nsvcs; i++) {
        printf("\t%s\n", svcs[i].svc_name);
    }

    exit(0);
}

/*
 * reload:
 * called on SIGHUP; reloads service database.
 */
static int
reload()
{
    if(debugmode) {
        fprintf(stderr, "Reloading service database.\n");
    }

    svc_clear();
    rc_loadall();
    signal(SIGHUP, reload);
}

/*
 * getargs:
 * construct an argv-style argument vector from an input string
 */
static char **
getargs(char *line)
{
    register char *p = line;
    int i = 0, nargs = 1, ws;
    char **argv;

    while(*p) {
	if(*p == ' ' || *p == '\t' || *p == '\n' || *p == '\r') {
	    nargs++;
	}
	p++;
    }

    argv = (char **) malloc(sizeof(char *) * (nargs + 1));

    p = line;
    argv[0] = p;
    while(*p) {
	/* skip whitespace */
	while(*p == ' ' || *p == '\t' || *p == '\n' || *p == '\r') {
	    *p = '\0';
	    ws = 1;
	    p++;
        }

        /* done with whitespace, found a new argument; store it */
	if(*p && ws) {
	    argv[++i] = p;
	    ws = 0;
        }

	p++;
    }

    argv[++i] = NULL;

    return argv;
}
